﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

namespace midi
{
    public static class ChordMaster
    {


        public static int[] ParseChord(string chord)
        {
            var re = new Regex(@"(?<note>[a-gA-G])(?<steps>[+-]?\d\d?(,[+-]?\d\d?)*)?", RegexOptions.Compiled);
            var match = re.Match(chord);
            var note = match.Groups["note"].Value;
            var steps = match.Groups["steps"].Value.Split(',');
            var intSteps = steps.Where(s => s != "").Select(x => Math.Abs(int.Parse(x)));
            var result = new List<int>() { CalculateNote0(note[0]) };
            Action<int, int> addStep = (step, norm) =>
            {
                var ss = steps.FirstOrDefault(s => s == "+" + step.ToString() || s == "-" + step.ToString() || s == step.ToString());
                if (ss == null)
                {
                    result.Add(result[0] + norm);
                }
                else
                {
                    switch (ss.Substring(0, 1))
                    {
                        case "+": result.Add(result[0] + norm + 1); break;
                        case "-": result.Add(result[0] + norm - 1); break;
                        default: result.Add(result[0] + norm); break;
                    }
                }
            };

            addStep(3, char.IsLower(note[0]) ? 3 : 4);
            addStep(5, 7);
            var maxStep = intSteps.Count() == 0 ? 5 : intSteps.Max();
            var restSteps = "6:9 7:11 9:14 11:17 13:19".Split(' ').ToDictionary(s => int.Parse(s.Split(':')[0]), s => int.Parse(s.Split(':')[1]));
            foreach (int rs in restSteps.Keys)
            {
                if (rs > maxStep) break;
                if (intSteps.Contains(rs))
                    addStep(rs, restSteps[rs]);
            }
            return result.ToArray();
        }

        public static int[] ParseChords(string chords, int fullDuration, int velocity, int channel)
        {
            var re = new Regex(@"(?<octave>\d)(?<note>[a-g])(?<alter>[\+\-]?)\.(?<template>\w+)", RegexOptions.Compiled);
            Func<Match, int[]> getmsgs = match =>
            {
                var note = CalculateNote0(match.Groups["note"].Value[0]) + int.Parse(match.Groups["octave"].Value) * 12 + (match.Groups["alter"].Value == "+" ? 1 : -1);
                var template = Properties.Resources.ResourceManager.GetString("ChordTemplate_" + match.Groups["template"].Value);
                return CreateChordFromTemplate(template, note, fullDuration, velocity, channel);
            };
            return (from Match m in re.Matches(chords) select m).SelectMany(getmsgs).ToArray();
        }

        public static int[] CreateChordFromTemplate(string template, int prima, int fullNoteDuration, int velocity, int channel)
        {
            var re = new Regex(@"(?<position>(\-?\d+)?)\.(?<duration>\d+(,\d+)*)", RegexOptions.Compiled);

            Func<string, int[]> melodyFromLine = line =>
            {
                var matches = re.Matches(line);
                var res = new List<int>();
                foreach (Match m in matches)
                {
                    var position = m.Groups["position"].Value;
                    var duration = m.Groups["duration"].Value.Split(',').Select(d => fullNoteDuration / int.Parse(d)).Sum();
                    if (position == "")
                    {
                        res.Add(0);
                        res.Add(duration);
                    }
                    else
                    {
                        var note = prima + int.Parse(position);
                        res.Add(velocity << 16 | note << 8 | 0x90 | channel);
                        res.Add(0);
                        res.Add(duration);
                        res.Add(note << 8 | 0x80 | channel);
                    }
                }
                return res.ToArray();
            };

            var lines = template.Split("\r\n".ToCharArray(), StringSplitOptions.RemoveEmptyEntries).Select(melodyFromLine).ToArray();
            return MessageProcessor.MergeMessagesLists(lines);
        }

        public static int[] CreateChordFromTemplate(string template, int[] notes, int fullNoteDuration, int velocity, int channel)
        {
            var re = new Regex(@"(?<position>(\-?\d+)?)\.(?<duration>\d+(,\d+)*)", RegexOptions.Compiled);

            Func<string, int[]> melodyFromLine = line =>
            {
                var matches = re.Matches(line);
                var res = new List<int>();
                foreach (Match m in matches)
                {
                    var position = m.Groups["position"].Value;
                    var duration = m.Groups["duration"].Value.Split(',').Select(d => fullNoteDuration / int.Parse(d)).Sum();
                    if (position == "")
                    {
                        res.Add(0);
                        res.Add(duration);
                    }
                    else
                    {
                        var note = notes[int.Parse(position)];
                        res.Add(velocity << 16 | note << 8 | 0x90 | channel);
                        res.Add(0);
                        res.Add(duration);
                        res.Add(note << 8 | 0x80 | channel);
                    }
                }
                return res.ToArray();
            };

            var lines = template.Split("\r\n".ToCharArray(), StringSplitOptions.RemoveEmptyEntries).Select(melodyFromLine).ToArray();
            return MessageProcessor.MergeMessagesLists(lines);
        }


        public static int[] GetChord(string chord)
        {
            var velocity = 64;
            var full = 3000;
            var prima = CalculateNote3(chord[0]) + 12;
            if (chord.Length > 1)
            {
                switch (chord[1])
                {
                    case '+': prima++; break;
                    case '-': prima--; break;
                    default:
                        break;
                }
            }
            var maj = char.IsUpper(chord[0]);
            var b2 = prima - 5;
            var ter = prima + (maj ? 4 : 3);
            var qu = prima + 7;
            var oct = prima + 12;
            var result = new List<int>();
            result.Add(velocity << 16 | prima << 8 | 0x91);
            result.Add(0);
            result.Add(full / 8);
            result.Add(velocity << 16 | ter << 8 | 0x91);
            result.Add(velocity << 16 | qu << 8 | 0x91);
            result.Add(velocity << 16 | oct << 8 | 0x91);
            result.Add(0);
            result.Add(full / 16);
            result.Add(ter << 8 | 0x81);
            result.Add(qu << 8 | 0x81);
            result.Add(oct << 8 | 0x81);
            result.Add(velocity << 16 | ter << 8 | 0x91);
            result.Add(velocity << 16 | qu << 8 | 0x91);
            result.Add(velocity << 16 | oct << 8 | 0x91);
            result.Add(0);
            result.Add(full / 16);
            result.Add(ter << 8 | 0x81);
            result.Add(qu << 8 | 0x81);
            result.Add(oct << 8 | 0x81);
            result.Add(prima << 8 | 0x81);

            result.Add(velocity << 16 | b2 << 8 | 0x91);
            result.Add(0);
            result.Add(full / 8);
            result.Add(velocity << 16 | ter << 8 | 0x91);
            result.Add(velocity << 16 | qu << 8 | 0x91);
            result.Add(velocity << 16 | oct << 8 | 0x91);
            result.Add(0);
            result.Add(full / 8);
            result.Add(ter << 8 | 0x81);
            result.Add(qu << 8 | 0x81);
            result.Add(oct << 8 | 0x81);
            result.Add(b2 << 8 | 0x81);
            return result.ToArray();
        }

        static int CalculateNote3(char n)
        {
            switch (char.ToLower(n))
            {
                case 'a': return 45;
                case 'b': return 47;
                case 'c': return 48;
                case 'd': return 50;
                case 'e': return 52;
                case 'f': return 53;
                case 'g': return 55;
                default:
                    throw new ArgumentException();
            }
        }

        static int CalculateNote0(char n)
        {
            switch (char.ToLower(n))
            {
                case 'a': return 9;
                case 'b': return 11;
                case 'c': return 0;
                case 'd': return 2;
                case 'e': return 4;
                case 'f': return 5;
                case 'g': return 7;
                default:
                    throw new ArgumentException();
            }
        }
    }
}
