﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

namespace midi
{
    class MusicFileParser
    {
        Dictionary<char, int> currentAlteratives = new Dictionary<char, int>()
        {
            {'a', 0}, {'b', 0}, {'c', 0}, {'d', 0}, {'e', 0}, {'f', 0}, {'g', 0}
        };

        Dictionary<char, int> notePosition = new Dictionary<char, int>()
        {
            {'a', 0}, {'b', 2}, {'c', -9}, {'d', -7}, {'e', -5}, {'f', -4}, {'g', -2}
        };

        const string sharpChars = "fcgdaeb";
        const string flatChars = "beadgcf";
        string key = "0";
        double fullNoteDuration = 4000.0;
        Regex noteAtom = new Regex(@"(?<octave>[0-5])(?<note>[a-g])(?<alterative>\+\+|--|\+|-|=)?(?<duration>\d+(,\d+)*)|(?<pause>x)(?<duration>\d+(,\d+)*)|(?<beat>\|)", RegexOptions.Compiled);
        Regex meta = new Regex(@"^(?<key>0|\d[+-])\s(?<fulldur>\d+)", RegexOptions.Compiled);

        void RefreshKey()
        {
            if (key == "0")
            {
                currentAlteratives.Keys.ToList().ForEach(c => currentAlteratives[c] = 0);
            }
            else if (key.Length == 2 && char.IsDigit(key[0]) && "+-".Contains(key[1]))
            {
                var plusOrMinus = key[1] == '+' ? 1 : -1;
                var alters = (plusOrMinus == 1 ? sharpChars : flatChars).Substring(0, key[0] - 48);
                currentAlteratives.Keys.ToList().ForEach(c => currentAlteratives[c] = alters.Contains(c) ? plusOrMinus : 0);
            }
            else
            {
                throw new FormatException("Строка, обозначающая тональность должна иметь значение \"0\" или состоять из двух символов, первый из которых - число, а второй - плюс или минус.");
            }
        }

        public string Parse(string notes)
        {
            var metaResult = meta.Match(notes);
            key = metaResult.Groups["key"].Value;
            fullNoteDuration = double.Parse(metaResult.Groups["fulldur"].Value);
            RefreshKey();
            var resultList = new List<string>();
            var noteMatches = noteAtom.Matches(notes);
            foreach (Match m in noteMatches)
            {
                if (m.Groups["beat"].Success) RefreshKey();
                else if (m.Groups["pause"].Success)
                {
                    var duration = CalculateDuration(m.Groups["duration"].Value);
                    resultList.Add("0");
                    resultList.Add(duration.ToString());
                }
                else
                {
                    var friquency = CalculateFriquency(m.Groups["octave"].Value, m.Groups["note"].Value, m.Groups["alterative"].Value);
                    var duration = CalculateDuration(m.Groups["duration"].Value);
                    resultList.Add(friquency.ToString());
                    resultList.Add(duration.ToString());
                }
            }

            return string.Join(" ", resultList);
        }

        int CalculateFriquency(string octave, string note, string alterative)
        {
            char noteChar = note[0];
            if (alterative.Length > 0)
            {
                int alt = 0;
                switch (alterative)
                {
                    case "+": alt = 1; break;
                    case "++": alt = 2; break;
                    case "-": alt = -1; break;
                    case "--": alt = -2; break;
                    case "=": alt = 0; break;
                    default:
                        break;
                }
                currentAlteratives[noteChar] = alt;
            }
            var distance = (double)((int.Parse(octave) - 1) * 12 + notePosition[noteChar] + currentAlteratives[noteChar]);
            double friq = 440.0 * Math.Pow(2.0, distance / 12.0);
            return (int)Math.Round(friq);
        }

        int CalculateDuration(string noteDur)
        {
            var dblResult = noteDur.Split(',').Select(d => fullNoteDuration / double.Parse(d)).Sum();
            return (int)Math.Round(dblResult);
        }
    }
}
